import numpy as np
from scipy.ndimage import map_coordinates

from vispy.scene.visuals import Compound, Markers, Rectangle, Ellipse, Line, Arrow, Polygon
from vispy.visuals.transforms import MatrixTransform, linear


# Compound from vispy.scene.visuals
class ControlPoints(Compound):
    def __init__(self, parent):
        Compound.__init__(self, [])
        self.unfreeze()
        self.parent = parent
        self._center = np.array([0, 0], dtype=np.float64)
        self._width = 0.0
        self._height = 0.0
        self._angle = 0.0  # in radians
        self.selected_cp = None
        self.opposed_cp = None

        self.edge_color = "black"
        self.face_color = (1, 1, 1, 0.2)
        self.marker_size = 8

        # Markers from vispy.scene.visuals
        self.control_points = [Markers(parent=self) for i in range(0, 4)]
        self.coords = np.zeros((len(self.control_points), 1, 2),
                               dtype=np.float32)

        for i, c in enumerate(self.control_points):
            c.set_data(
                pos=self.coords[i],
                edge_color=self.edge_color,
                face_color=self.face_color,
                size=self.marker_size,
            )
            c.interactive = True

        self.transform = linear.STTransform(translate=(0, 0, -2))

        self.freeze()

    def update_bounds(self):
        self._center = np.array(
            [
                0.5 * (self.parent.bounds(0)[1] + self.parent.bounds(0)[0]),
                0.5 * (self.parent.bounds(1)[1] + self.parent.bounds(1)[0]),
            ],
            dtype=np.float64,
        )
        self._width = abs(self.parent.bounds(0)[1] - self.parent.bounds(0)[0])
        self._height = abs(self.parent.bounds(1)[1] - self.parent.bounds(1)[0])
        self.update_points()

    def update_points(self):
        # center-+
        self.coords[0] = np.array(
            [
                [
                    self._center[0]
                    - 0.5 * np.cos(self._angle) * self._width
                    + 0.5 * np.sin(self._angle) * self._height,
                    self._center[1]
                    + 0.5 * np.sin(self._angle) * self._width
                    + 0.5 * np.cos(self._angle) * self._height,
                ]
            ]
        )
        # center++
        self.coords[1] = np.array(
            [
                [
                    self._center[0]
                    + 0.5 * np.cos(self._angle) * self._width
                    + 0.5 * np.sin(self._angle) * self._height,
                    self._center[1]
                    - 0.5 * np.sin(self._angle) * self._width
                    + 0.5 * np.cos(self._angle) * self._height,
                ]
            ]
        )
        # center+-
        self.coords[2] = np.array(
            [
                [
                    self._center[0]
                    + 0.5 * np.cos(self._angle) * self._width
                    - 0.5 * np.sin(self._angle) * self._height,
                    self._center[1]
                    - 0.5 * np.sin(self._angle) * self._width
                    - 0.5 * np.cos(self._angle) * self._height,
                ]
            ]
        )
        # center--
        self.coords[3] = np.array(
            [
                [
                    self._center[0]
                    - 0.5 * np.cos(self._angle) * self._width
                    - 0.5 * np.sin(self._angle) * self._height,
                    self._center[1]
                    + 0.5 * np.sin(self._angle) * self._width
                    - 0.5 * np.cos(self._angle) * self._height,
                ]
            ]
        )
        for i, c in enumerate(self.control_points):
            c.set_data(
                pos=self.coords[i],
                edge_color=self.edge_color,
                face_color=self.face_color,
                size=self.marker_size,
            )

    def select(self, val, obj=None):
        self.visible(val)
        self.selected_cp = None
        self.opposed_cp = None

        if obj is not None:
            n_cp = len(self.control_points)
            for i in range(0, n_cp):
                c = self.control_points[i]
                if c == obj:
                    self.selected_cp = c
                    self.opposed_cp = self.control_points[int((i + n_cp / 2)) % n_cp]

    def start_move(self, start):
        self.parent.start_move(start)

    def move(self, end, modifiers=[], *args, **kwargs):
        if not self.parent.editable:
            return
        if self.selected_cp is not None:

            # alt for scaling from center mode
            if "Alt" in modifiers:
                diag = end - self._center

                self._width = 2 * abs(
                    np.cos(self._angle) * diag[0] - np.sin(self._angle) * diag[1]
                )
                self._height = 2 * abs(
                    np.sin(self._angle) * diag[0] + np.cos(self._angle) * diag[1]
                )
                if "Control" in modifiers:
                    val = max(self._width, self._height)
                    self._width = val
                    self._height = val

                self.update_points()
                self.parent.update_from_controlpoints()

            # normal scaling mode where one corner is fixed
            else:
                opp_index = self.control_points.index(self.opposed_cp)

                opp = self.coords[opp_index, 0, :]
                diag = end - opp
                center = opp + 0.5 * diag

                self._width = (
                    np.cos(self._angle) * diag[0] - np.sin(self._angle) * diag[1]
                )
                self._height = (
                    np.sin(self._angle) * diag[0] + np.cos(self._angle) * diag[1]
                )

                # depending on selected point
                if opp_index == 2 or opp_index == 1:
                    self._width *= -1
                if opp_index == 0 or opp_index == 1:
                    self._height *= -1

                # square rectangle when holding Ctrl
                if "Control" in modifiers:
                    val = np.max(np.abs([self._width, self._height]))
                    self._width = np.sign(self._width) * val
                    self._height = np.sign(self._height) * val

                    # I have forgotten why this works
                    if opp_index == 2 or opp_index == 0:
                        offset = (
                            val
                            / np.sqrt(2)
                            * np.array(
                                [
                                    np.sign(self._width)
                                    * np.cos(
                                        np.pi / 4
                                        + np.sign(self._width)
                                        * np.sign(self._height)
                                        * self.angle
                                    ),
                                    np.sign(self._height)
                                    * np.sin(
                                        np.pi / 4
                                        + np.sign(self._width)
                                        * np.sign(self._height)
                                        * self.angle
                                    ),
                                ]
                            )
                        )
                    else:
                        offset = (
                            val
                            / np.sqrt(2)
                            * np.array(
                                [
                                    np.sign(self._width)
                                    * np.cos(
                                        np.pi / 4
                                        - np.sign(self._width)
                                        * np.sign(self._height)
                                        * self.angle
                                    ),
                                    np.sign(self._height)
                                    * np.sin(
                                        np.pi / 4
                                        - np.sign(self._width)
                                        * np.sign(self._height)
                                        * self.angle
                                    ),
                                ]
                            )
                        )

                    if opp_index == 2 or opp_index == 1:
                        offset[0] *= -1
                    if opp_index == 0 or opp_index == 1:
                        offset[1] *= -1
                    center = opp + offset

                self.set_center(center)
                self.parent.update_from_controlpoints()
                self.parent.update_transform()

    def rotate(self, angle):
        if self.parent.editable:
            # if self.control_points.index(self.selected_cp)%2 == 0:
            self._angle = angle - self.parent.drag_reference_angle
            self.update_points()
            self.parent.update_transform()

    def visible(self, v):
        for c in self.control_points:
            c.visible = v

    def get_center(self):
        return self._center

    @property
    def center(self):
        return self._center

    def set_center(self, val):
        self._center = val
        self.update_points()

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, val):
        self._angle = val
        self.update_points()

    def delete(self):
        self.parent.delete()
        del self

    def get_coords(self):
        return self.coords[:, 0, :].copy()


# from vispy.scene.visuals
class EditVisual(Compound):

    def __init__(
        self,
        settings,
        control_points=None,
        editable=True,
        selectable=True,
        on_select_callback=None,
        callback_argument=None,
        angle=0,
        *args,
        **kwargs,
    ):

        # from vispy.scene.visuals
        Compound.__init__(self, [], *args, **kwargs)
        self.unfreeze()
        self.editable = editable
        self._selectable = selectable
        self.form = None
        self._on_select_callback = on_select_callback
        self._callback_argument = callback_argument
        self.settings = settings

        match control_points:
            case ("LineControlPoints", int(), _):
                _, num_points, coords = control_points
                self.control_points = LineControlPoints(
                    parent=self, num_points=num_points, coords=coords
                )
            case _:
                self.control_points = ControlPoints(parent=self)
                self.angle = angle

        self.drag_reference = np.array([0, 0], dtype=np.float64)
        self.drag_reference_angle = 0.0

        self.transform = linear.STTransform(translate=(0, 0, -1))
        self.freeze()

    def add_subvisual(self, visual):
        # from vispy.scene.visuals
        Compound.add_subvisual(self, visual)

    def select(self, val, obj=None):
        if self.selectable:
            self.control_points.visible(val)
            if self._on_select_callback is not None:
                self._on_select_callback(self._callback_argument)

    @property
    def selected(self):
        return self.control_points.control_points[0].visible

    def start_move(self, start):
        self.drag_reference = start[0:2] - self.control_points.get_center()

        # calculate angle of drag reference point
        x, y = self.drag_reference
        # -y because y-axis is flipped
        angle_drag = np.arctan2(-y, x)
        self.drag_reference_angle = angle_drag - self.control_points.angle

    def move(self, end, *args, **kwargs):
        if self.editable:
            shift = end[0:2] - self.drag_reference
            self.set_center(shift)

            self.update_transform()

    def rotate(self, angle):
        """
        Rotate the object by angle (in radians) around its center from the drag reference point.
        """

        if self.editable:

            # shift is difference
            abs_angle = angle - self.drag_reference_angle
            # rotate control points and therefore shape as well
            self.angle = abs_angle % (2 * np.pi)
            self.update_transform()

    def update_transform(self):
        tr = MatrixTransform()
        tr.translate(-self.center)
        tr.rotate(np.rad2deg(self.angle), (0, 0, -1))
        tr.translate(self.center)
        self.form.transform = tr

    def update_from_controlpoints(self):
        pass

    @property
    def selectable(self):
        return self._selectable

    @selectable.setter
    def selectable(self, val):
        self._selectable = val

    def set_visibility(self, v):
        self.visible = v
        # only make controlpoint not visible but never visible
        # that part is already done by the clicking architecture
        if not v:
            self.control_points.visible(v)

    @property
    def angle(self):
        return self.control_points.angle

    @angle.setter
    def angle(self, val):
        self.control_points.angle = val
        # self.update_from_controlpoints()

    # height
    @property
    def height(self):
        return self.control_points._height

    @height.setter
    def height(self, val):
        self.control_points._height = val
        self.control_points.update_points()
        self.update_from_controlpoints()

    # width
    @property
    def width(self):
        return self.control_points._width

    @width.setter
    def width(self, val):
        self.control_points._width = val
        self.control_points.update_points()
        self.update_from_controlpoints()

    # center
    @property
    def center(self):
        return self.control_points.get_center()

    @center.setter
    # this method redirects to set_center. Override set_center in subclasses.
    def center(self, val):
        self.set_center(val)

    # override this method in subclass
    def set_center(self, val):
        self.control_points.set_center(val[0:2])

    def select_creation_controlpoint(self):
        self.control_points.select(True, self.control_points.control_points[1])

    def output_properties(self):
        None

    def delete(self):
        self.parent = None

    def update_colors(self, color, border_color):
        pass

    def update_property(self, prop, val, scaling_factor=None):
        pass

    def get_modifiable_properties(self):
        return None

class EditRectVisual(EditVisual):
    def __init__(
        self,
        center=np.array([0, 0], dtype=np.float64),
        width=1e-6,
        height=1e-6,
        border_width=2,
        *args,
        **kwargs,
    ):

        EditVisual.__init__(self, *args, **kwargs)
        self.unfreeze()

        
        arrow_color = self.settings.value("graphics/object_border_color").getRgb()
        border_color = tuple([value / 255 for value in arrow_color]),
        self.arrow = Arrow(
            pos=np.array([center - np.array([width / 3, 0]), center + np.array([width / 3, 0])]),
            color=border_color,
            arrow_size=10,
            method='gl',
            width=5,
            antialias=True,
            arrow_type='stealth',
            arrows=np.array([[center[0] - width / 3, center[1] , center[0] + width / 3, center[1]]]),
            arrow_color=border_color,
            parent=self,
        )
        
        self.form = Rectangle(
            center=center,
            width=width,
            height=height,
            border_width=border_width,
            radius=0,
            parent=self,
        )
        self.form.interactive = True
        
        self.arrow.visible = False

        color = self.settings.value("graphics/object_color").getRgb()
        self.form.color = tuple([value / 255 for value in color])
        self.form.border_color = border_color

        self.freeze()
        self.add_subvisual(self.form)
        self.add_subvisual(self.arrow)
        self.control_points.update_bounds()
        self.rotate(self.angle)
        
    def move(self, end, *args, **kwargs):
        super().move(end, *args, **kwargs)
        # arrow properties
        self.update_arrow()
        
    def hide_arrow(self):
        self.arrow.visible = False
        
    def show_arrow(self):
        self.arrow.visible = True
        
    def update_arrow(self):
        self.arrow.set_data(pos=np.array([self.center - np.array([self.width / 3, 0]), self.center + np.array([self.width / 3, 0])]),
                            arrows=np.array([[self.center[0] - self.width / 3, self.center[1], self.center[0] + self.width / 3, self.center[1]]]))
        
    def set_center(self, val):
        self.control_points.set_center(val[0:2])
        self.form.center = val[0:2]

    def update_from_controlpoints(self):
        try:
            self.form.width = abs(self.control_points._width)
        except ValueError:
            None
        try:
            self.form.height = abs(self.control_points._height)
        except ValueError:
            None
        try:
            self.form.center = self.control_points.center
        except ValueError:
            None
        try:
            self.angle = self.control_points.angle
            self.update_transform()
        except ValueError:
            None
        self.update_arrow()

    def output_properties(self):

        return dict(
            center=(self.form.center, "px"),
            width=(self.form.width, "px"),
            height=(self.form.height, "px"),
            area=(self.form.height * self.form.width, "px²"),
            angle=(np.rad2deg(self.control_points._angle), "°"),
        )

    def update_property(self, prop, val, scaling_factor=None):
        if scaling_factor is None:
            scaling_factor = 1
        match prop:
            case "center":
                self.set_center(val / scaling_factor)
            case "width":
                self.width = val / scaling_factor
            case "height":
                self.height = val / scaling_factor
            case "angle":
                self.angle = np.deg2rad(val)
                self.update_transform()
    
    def get_modifiable_properties(self):
        return ["center", "width", "height", "angle"]

    def save(self):
        return dict(
            center=self.form.center,
            width=self.form.width,
            height=self.form.height,
            angle=self.angle,
        )

    def intensity_profile(self, image, n_x=100, n_y=20, **kwargs):
        # Get the coordinates of the rectangle
        coords = self.control_points.coords[:, 0, :]

        # Compute the vector along y-direction
        x_dir = np.linspace(0, coords[3][0] - coords[0][0], n_y, endpoint=True)
        y_dir = np.linspace(0, coords[3][1] - coords[0][1], n_y, endpoint=True)
        # Compute the start and end points for all lines
        starts = coords[0][::-1] + np.column_stack((y_dir, x_dir))
        ends = coords[1][::-1] + np.column_stack((y_dir, x_dir))

        if self.control_points._width < 0:
            starts, ends = ends, starts

        # Initialize an empty list to hold the intensity profiles
        intensity_profile_line = np.zeros(n_x)
        evaluation_coords = np.empty((n_y, n_x, 2))
        img = np.sum(image, axis=2)
        # Iterate over all lines
        for i, (start, end) in enumerate(zip(starts, ends)):
            # Compute the coordinates along the line
            x_coords = np.linspace(start[1], end[1], n_x)
            y_coords = np.linspace(start[0], end[0], n_x)

            evaluation_coords[i, :, :] = np.column_stack((x_coords, y_coords))
            # Stack the coordinates into a 2D array
            eval_coords = np.vstack([y_coords, x_coords])

            # Compute the intensity profile for the line
            intensity_profile_line += map_coordinates(
                img, eval_coords, mode="constant", **kwargs
            )

        return intensity_profile_line, evaluation_coords

    def update_colors(self, color, border_color):
        self.form.color = color
        self.form.border_color = border_color
        self.arrow.color = border_color
        self.arrow.arrow_color = border_color


class EditEllipseVisual(EditVisual):
    """ Visual representation of an ellipse object with control points.
    """
    
    def __init__(
        self,
        center=np.array([0, 0], dtype=np.float64),
        radius=np.array([1e-6, 1e-6], dtype=np.float64),
        *args,
        **kwargs,
    ):

        EditVisual.__init__(self, *args, **kwargs)
        self.unfreeze()

        self.form = Ellipse(center=center, radius=radius, border_width=2, parent=self)

        color = self.settings.value("graphics/object_color").getRgb()
        self.form.color = tuple([value / 255 for value in color])
        border_color = self.settings.value("graphics/object_border_color").getRgb()
        self.form.border_color = tuple([value / 255 for value in border_color])

        self.form.interactive = True

        self.freeze()
        self.add_subvisual(self.form)
        self.control_points.update_bounds()
        self.rotate(self.angle)

    def set_center(self, val):
        self.control_points.set_center(val)
        self.form.center = val

    def update_from_controlpoints(self):
        try:
            self.form.radius = np.array(
                [
                    0.5 * abs(self.control_points._width),
                    0.5 * abs(self.control_points._height),
                ],
                dtype=np.float64,
            )
        except ValueError:
            None
        try:
            self.form.center = self.control_points.center
        except ValueError:
            None
        try:
            self.angle = self.control_points.angle
            self.update_transform()
        except ValueError:
            None

    def output_properties(self):

        if self.form.radius[0] == self.form.radius[1]:
            radius = self.form.radius[0]
        else:
            radius = self.form.radius
        return dict(
            center=(self.form.center, "px"),
            radius=(radius, "px"),
            area=(np.prod(self.form.radius) * np.pi, "px²"),
            angle=(np.rad2deg(self.control_points._angle), "°"),
        )

    def update_property(self, prop, val, scaling_factor=None):
        if scaling_factor is None:
            scaling_factor = 1
        match prop:
            case "center":
                self.set_center(val / scaling_factor)
            case "radius":
                self.width = val[0] / scaling_factor
                self.height = val[1] / scaling_factor
            case "angle":
                self.angle = np.deg2rad(val)
                self.update_transform()
    
    def get_modifiable_properties(self):
        return ["center", "radius", "angle"]

    def save(self):
        return dict(
            center=self.form.center,
            radius=np.array(self.form.radius),
            angle=self.angle,
        )

    def update_colors(self, color, border_color):
        self.form.color = color
        self.form.border_color = border_color


class LineControlPoints(Compound):

    continue_adding_points: bool = True
    edge_color:str = "black"
    face_color:tuple = (1, 1, 1, 0.2)
    marker_size:int = 8
    _length:float = 0.0

    def __init__(self, parent, num_points=2, coords=None, *args, **kwargs):
        Compound.__init__(self, [], *args, **kwargs)
        self.unfreeze()
        self.parent = parent
        self.num_points:int = num_points
        
        # if numpoints = 0 we can attach points but start with a simple line
        if coords is not None:
            self.coords = coords
            self.continue_adding_points = False
        else:
            self.coords = np.zeros((1, 2), dtype=np.float64)
            
        self.selected_cp = None

        self.control_points = [Markers(parent=self) for i in range(0, len(self.coords))]
        for cpoint, coord in zip(self.control_points, self.coords):
            cpoint.set_data(
                pos=np.array([coord], dtype=np.float32),
                edge_color=self.edge_color,
                face_color=self.face_color,
                size=self.marker_size,
            )
            cpoint.interactive = True
        
        self.transform = linear.STTransform(translate=(0, 0, -2))

        self.freeze()

    @property
    def length(self):
        diff = np.diff(self.coords, axis=0)
        self._length = np.abs(np.linalg.norm(diff, axis=1))
        if len(self._length) == 1:
            self._length = self._length[0]
        return self._length
    
    def add_point(self, point: np.ndarray, index=None, select=True, show=True):
        """Adds a new control point to the list of control points at index, 
        or after selected_cp if no index is given."""
        
        if (len(self.control_points) < self.num_points 
            or self.num_points == 0):
            # index of current cp to insert new point after it
            if index is None:
                index = self.get_selected_index()+1
            self.coords = np.vstack((self.coords[:index], point, self.coords[index:]))
            c_point = Markers(parent=self,
                                pos=np.array([point], dtype=np.float32),
                                edge_color=self.edge_color,
                                face_color=self.face_color,
                                size=self.marker_size,
            )
            c_point.interactive = True
            self.control_points.insert(index, c_point)
            
            # make the new marker the selected
            if select:
                self.select(True, c_point)
            else:
                c_point.visible = show

        else:
            # self.coords[-1] = point
            self.continue_adding_points = False   
        
        self.update_points()
        self.parent.update_from_controlpoints()

    def remove_point(self, index=None):
        """Removes point at index from the list of control points
        or the selected point if no index is given.
        """
        # select the proper control point by index
        if index is not None:
            self.selected_cp = self.control_points[index]
        # check if there are more than 2 control points
        if self.selected_cp is not None and len(self.control_points) > 2:
            index = self.get_selected_index()
            self.selected_cp.parent = None
            self.selected_cp = self.control_points[index - 1]
            # remove control point and corresponding coordinate
            del self.control_points[index]
            self.coords = np.delete(self.coords, index, axis=0)
        
        self.update_points()
        self.parent.update_from_controlpoints()

    def get_selected_index(self):
        return self.control_points.index(self.selected_cp)

    def update_points(self):

        for cpoint, coord in zip(self.control_points, self.coords):
            cpoint.set_data(
                pos=np.array([coord]),
                edge_color=self.edge_color,
                face_color=self.face_color,
                size=self.marker_size,
            )

    def select(self, val, obj=None):
        self.visible(val)
        self.selected_cp = None

        if obj is not None:
            index = self.control_points.index(obj)
            if index != -1:
                self.selected_cp = self.control_points[index]

    def start_move(self, start):
        self.parent.start_move(start)

    def move(self, end, *args, **kwargs):

        if not self.parent.editable:
            return
        if self.selected_cp is not None:
            index = self.control_points.index(self.selected_cp)
            self.coords[index] = end[0:2]

            self.update_points()
            self.parent.update_from_controlpoints()

    def visible(self, v):
        for c in self.control_points:
            c.visible = v

    def set_coords(self, coords):
        self.coords = coords
        self.update_points()

    def get_coords(self):
        return self.coords.copy()
    
    def get_center(self):
        return np.array(self.coords[0]) + 0.5 * (
            np.array(self.coords[-1]) - np.array(self.coords[0])
        )

    def set_center(self, val):
        shift = val - self.get_center()
        self.coords += shift

        self.update_points()

    def delete(self):
        self.parent.delete()


class EditLineVisual(EditVisual):

    def __init__(self, num_points=2, coords=None, *args, **kwargs):

        EditVisual.__init__(
            self, control_points=("LineControlPoints", num_points, coords), *args, **kwargs
        )
        self.unfreeze()

        self.num_points = num_points
        border_color = self.settings.value("graphics/object_border_color").getRgb()
        self.line_color = tuple([value / 255 for value in border_color])
        self.line_width = 3

        # this is to allow loading of structures
        if coords is not None:
            self.coords = coords

        self.form = Line(
            pos=self.coords,
            width=self.line_width,
            color=self.line_color,
            method="gl",
            antialias=True,
            parent=self,
        )

        self.form.interactive = True

        self.freeze()
        self.add_subvisual(self.form)

    @property
    def length(self):
        return self.control_points.length

    @property
    def angles(self):
        diff = np.diff(self.control_points.coords, axis=0)
        diff[0] *= -1
        angles = np.rad2deg(np.arctan2(-diff[:, 1], diff[:, 0]))
        return angles

    @property
    def angle(self):
        return self.angles[0]

    def start_move(self, start):
        self.drag_reference = start[:2] - self.control_points.get_center()

    @property
    def coords(self):
        return self.control_points.coords

    @coords.setter
    def coords(self, coords):
        self.control_points.set_coords(coords)

    def move(self, end, *args, **kwargs):
        if self.editable:
            new_center = end[0:2] - self.drag_reference
            self.set_center(new_center)
            self.update_from_controlpoints()

    def rotate(self, angle):
        None

    def update_transform(self):
        None

    def update_from_controlpoints(self):
        try:
            self.form.set_data(
                pos=self.coords, width=self.line_width, color=self.line_color
            )
        except ValueError:
            None

    @property
    def selectable(self):
        return self._selectable

    @selectable.setter
    def selectable(self, val):
        self._selectable = val

    def select_creation_controlpoint(self):
        self.control_points.select(True, self.control_points.control_points[0])

    def output_properties(self):
        if len(self.control_points.control_points) < 2:
            return dict()
        if len(self.control_points.control_points) == 2:
            angle = self.angles
        else:
            angle = np.abs(np.diff(self.angles))
            for i, a in enumerate(angle):
                if a > 180:
                    angle[i] = 360 - a

        return dict(length=[self.length, "px"], angle=[angle[0], "°"])

    def update_property(self, prop, val, scaling_factor=None):
        if scaling_factor is None:
            scaling_factor = 1
        match prop:
            case "length":
                if isinstance(val, np.ndarray):
                    norm_vecs = (
                        np.diff(self.control_points.coords, axis=0) / self.length
                    )
                    for i, vec in enumerate(norm_vecs):
                        self.control_points.coords[i + 1] = (
                            self.control_points.coords[i]
                            + val[i] / scaling_factor * vec
                        )

                else:
                    # get normalized vector
                    norm_vec = np.diff(self.control_points.coords, axis=0) / self.length
                    self.control_points.coords[1] = (
                        self.control_points.coords[0] + val / scaling_factor * norm_vec
                    )
                self.update_from_controlpoints()
                self.control_points.update_points()
    
    def get_modifiable_properties(self):
        return ["length"]

    def save(self):
        return dict(coords=self.coords, num_points=self.num_points)

    def intensity_profile(self, image, n=100, **kwargs):

        # Iterate over pairs of consecutive coordinates

        diff = np.diff(self.control_points.coords, axis=0)
        lengths = np.abs(np.linalg.norm(diff, axis=1))

        # Compute the total length of the path
        total_length = np.sum(lengths)

        # Compute the number of points for each line segment
        num_points = np.round(n * lengths / total_length).astype(int)

        length = np.sum(num_points)
        intensity_profiles = np.empty(length)
        evaluation_coords = np.empty((length, 2))

        # Convert the image to grayscale
        img = np.sum(image, axis=2)

        count = 0
        for index, m in enumerate(num_points):
            # Swap x and y in the start and end points
            start = self.coords[index][::-1]
            end = self.coords[index + 1][::-1]

            # Compute the coordinates along the line
            x_coords = np.linspace(start[1], end[1], m, endpoint=True)
            y_coords = np.linspace(start[0], end[0], m, endpoint=True)

            # Stack the coordinates into a 2D array
            eval_coords = np.vstack([y_coords, x_coords])
            evaluation_coords[count : count + m] = eval_coords.T[:, ::-1]

            # Get the intensity profile along the line segment
            intensity_profile = map_coordinates(
                img, eval_coords, mode="constant", **kwargs
            )

            # Append the intensity profile to full array
            intensity_profiles[count : count + m] = intensity_profile

            count += m

        return intensity_profiles, evaluation_coords
    
class EditPolygonVisual(EditLineVisual):
    def __init__(self, num_points=0, coords=None, *args, **kwargs):

        EditLineVisual.__init__(self, 
                            num_points=num_points,
                            coords=coords,
                            *args, **kwargs)
        
        self.swap_form()
        
    def swap_form(self):
        # self.form.parent = None
        self.remove_subvisual(self.form)
        # self.form.parent = None
        if len(self.coords) < 3 and isinstance(self.form, Polygon):
            # remove old self.form
            # TODO: this is ugly as hell, as it is still floating around
            self.form.visible = False
            # but I dont know how to remove it properly
            
            self.form = Line(
                pos=self.coords,
                width=self.line_width,
                color=self.line_color,
                method="gl",
                antialias=True,
                parent=self,
            )        
                    
        elif len(self.coords) >= 3 and isinstance(self.form, Line):
            # remove old self.form
            # TODO: this is ugly as hell, as it is still floating around
            self.form.visible = False
            # but I dont know how to remove it properly
            self.form = Polygon(
                pos=self.corrected_coords(),
                border_method='gl',
                border_width=2,
                parent=self,
            )
            
            color = self.settings.value("graphics/object_color").getRgb()
            self.form.color = tuple([value / 255 for value in color])
            border_color = self.settings.value("graphics/object_border_color").getRgb()
            self.form.border_color = tuple([value / 255 for value in border_color])
        else:
            self.form.parent = None
            
        self.form.parent=self
        self.add_subvisual(self.form) 
        self.form.interactive = True
        
    def corrected_coords(self):
        """check if two consecutive points are the same and return a view without them"""
        mask = np.any(self.coords[1:] != self.coords[:-1], axis=1)
        corrected = self.coords[np.insert(mask, 0, True)]
        if len(corrected) > 2:
            return corrected
        return self.coords
    
    def update_from_controlpoints(self):
        self.swap_form()
        if len(self.coords) < 3:
            super().update_from_controlpoints()
        else:
            self.form.pos = self.corrected_coords()
                                    
    def intensity_profile(self, image, n=100, **kwargs):
        pass
        


from measury.app import App
from pathlib import Path
import pytest
from numpy import __version__ as np_version

if np_version[0] == "2":
    test_files = ["test_image.tif", "test_file.msry", "test_file_2.msry"]
else:
    test_files = ["test_image.tif", "test_file.msry", "test_file_3.msry", "test_file_4.msry"]

def test_main_program():
    
    app = App()
    app.run(run_vispy=False)
    app.vispy_app.process_events()
    # app.vispy_app.sleep(1)
    app.close()
    
def test_right_ui():
    
    app = App()
    app.run(run_vispy=False)
    app.vispy_app.process_events()
    app.main_window.right_ui.show_ui()
    app.vispy_app.process_events()
    app.main_window.right_ui.hide_ui()
    app.vispy_app.process_events()
    app.close()
    
@pytest.mark.parametrize("file", test_files)
def test_open_file(file):
    
    app = App()
    app.run(run_vispy=False)
    app.vispy_app.process_events()
    app.data_handler.open_file(Path(__file__).parent/"test_data"/file)
    app.close()

def test_identify_scaling():

    app = App()
    app.run(run_vispy=False)
    app.vispy_app.process_events()
    app.data_handler.open_file(Path(__file__).parent/Path("test_data/test_image.tif"))
    app.main_window.main_ui.automatic_scaling()
    app.close()

@pytest.mark.parametrize("file", test_files)
def test_start_with_image(file):

    app = App(file_path=Path(__file__).parent/"test_data"/file)
    app.run(run_vispy=False)
    app.vispy_app.process_events()
    app.close()
    
def test_open_right_ui():

    app = App(file_path=Path(__file__).parent/"test_data"/test_files[-1])
    app.run(run_vispy=False)
    app.vispy_app.process_events()
    app.main_window.right_ui.show_ui()
    app.vispy_app.process_events()
    app.main_window.right_ui.hide_ui()
    app.vispy_app.process_events()
    app.close()

def test_settings_page():

    app = App()
    app.run(run_vispy=False)
    app.vispy_app.process_events()
    app.main_window.open_settings_page()
    app.vispy_app.process_events()
    app.close()
    


from PySide6.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QLabel,
    QPushButton,
    QHBoxLayout,
    QGroupBox,
    QLineEdit,
    QComboBox,
)
from PySide6.QtGui import QIntValidator

import numpy as np
from vispy.scene import LinePlot, InfiniteLine, Text, SceneCanvas
from vispy.scene.widgets import AxisWidget, Label

# relative imports
from .drawable_objects import (
    EditRectVisual,
    EditLineVisual,
    EditPolygonVisual,
)


class RightUI(QWidget):

    content_size = None

    def __init__(self, main_window, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.main_window = main_window
        self.initUI()

    def initUI(self):
        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        self.top_layout = QHBoxLayout()
        self.layout.addLayout(self.top_layout)

        self.top_layout.addWidget(QLabel("Advanced Measurement Tools"))

        self.top_layout.addStretch()

        # button to close the right ui
        self.close_button = QPushButton("X")
        self.close_button.clicked.connect(self.hide_ui)
        self.close_button.setFixedSize(20, 20)
        self.close_button.setStyleSheet(
            "background-color: red; color: white; border: none;"
        )
        self.top_layout.addWidget(self.close_button)

        # add intensity plot
        self.vispy_intensity_plot = IntensityPlot(self.main_window)
        self.layout.addWidget(self.vispy_intensity_plot.native)

        # add buttons to customize interpolation
        self.interpolation_box = QGroupBox("Interpolation Settings")
        self.layout.addWidget(self.interpolation_box)
        self.interpolation_layout = QVBoxLayout()
        self.interpolation_box.setLayout(self.interpolation_layout)
        self.ppx_layout = QHBoxLayout()
        self.interpolation_layout.addLayout(self.ppx_layout)
        # add qlineedit that allows only integers
        self.ppx_edit = QLineEdit()
        self.ppx_edit.setValidator(QIntValidator(1, 1000))
        self.ppx_edit.setPlaceholderText("1")
        self.ppx_layout.addWidget(QLabel("points per pixel: "))
        self.ppx_layout.addWidget(self.ppx_edit)
        self.ppx_edit.textChanged.connect(self.update_intensity_plot)

        self.order_layout = QHBoxLayout()
        self.interpolation_layout.addLayout(self.order_layout)
        self.order_layout.addWidget(QLabel("Spline interpolation order: "))
        self.order_dd = QComboBox()
        self.order_dd.addItems([str(i) for i in range(6)])
        self.order_layout.addWidget(self.order_dd)
        self.order_dd.currentIndexChanged.connect(
            lambda: self.update_intensity_plot(resize=False)
        )

        self.save_button = QPushButton("Save Intensity Profile")
        self.save_button.clicked.connect(self.save_intensity_plot)
        self.layout.addWidget(self.save_button)
        # add spacer
        self.layout.addStretch()

    def resizeEvent(self, event):
        super().resizeEvent(event)

        if self.width() == 0:
            self.main_window.open_right_ui_button.show()
        else:
            self.main_window.open_right_ui_button.hide()

    def hide_ui(self):
        self.content_size = self.width()
        size = self.main_window.splitter.sizes()
        self.main_window.splitter.setSizes(
            [size[0] + self.content_size // 2,
             size[1] + self.content_size // 2,
             0]
        )

        self.hide()
        self.main_window.open_right_ui_button.show()
        self.main_window.vispy_canvas.hide_arrows()

    def show_ui(self):
        self.show()
        self.main_window.open_right_ui_button.hide()
        # self.main_window.vispy_canvas.show_arrows()

        self.update_intensity_plot()

        # resize layout
        size = self.main_window.splitter.sizes()
        if self.content_size is None:
            self.content_size = self.layout.sizeHint().width()
        self.main_window.splitter.setSizes(
            [
                size[0] - self.content_size // 2,
                size[1] - self.content_size // 2,
                self.content_size,
            ]
        )

    def reset_intensity_plot(self):
        self.vispy_intensity_plot.intensity_line.set_data([[0, 0], [0, 0]])
        self.vispy_intensity_plot.diagram.camera.set_range(x=(0, 1), y=(0, 1))

    def update_intensity_plot(self, resize=True):
        if self.isHidden():
            return
        
        # scaling factor for the intensity plot
        if self.main_window.main_ui.scaling_factor is None:
            scaling_factor = 1
            self.vispy_intensity_plot.xaxis_label.text = "distance (px)"
        else:
            scaling_factor = self.main_window.main_ui.scaling_factor
            self.vispy_intensity_plot.xaxis_label.text =  f"distance ({self.main_window.main_ui.units_dd.currentText()})"
            
        selected_element = self.main_window.vispy_canvas.get_selected_object()
        
        # show only the active arrow
        self.main_window.vispy_canvas.hide_arrows()
        if selected_element is not None and isinstance(selected_element, EditRectVisual):
            selected_element.show_arrow()
        
        if (not isinstance(selected_element, EditRectVisual) 
        and not type(selected_element) is EditLineVisual):
            self.reset_intensity_plot()
            return 
        
        image = self.main_window.data_handler.img_data
        # get ui settings
        interpolation_factor = (
            int(self.ppx_edit.text()) if self.ppx_edit.text() else 1
        )
        spline_order = int(self.order_dd.currentText())
        

        # calculate intensity profile
        if isinstance(selected_element, EditLineVisual):
            length = selected_element.length
            if not isinstance(length, float):
                length = np.sum(length)
            intensity, _ = selected_element.intensity_profile(
                image=image,
                n=interpolation_factor * int(length),
                order=spline_order,
            )
            distance = np.linspace(0, length, len(intensity))*scaling_factor
            if selected_element.angle % 90 == 0:
                distance += np.min(selected_element.control_points.coords[:, 0])*scaling_factor
        elif isinstance(selected_element, EditRectVisual):
            if spline_order > 1:
                spline_order = 1
                self.order_dd.setCurrentText("1")
                self.main_window.raise_error(
                    "Spline order for rectangles is limited to 1"
                )

            width = selected_element.width
            width_integer = abs(int(width))
            height_integer = abs(int(selected_element.height))
            if width_integer <= 0 or height_integer <= 0:
                self.reset_intensity_plot()
                return
            intensity, _ = selected_element.intensity_profile(
                image=image,
                n_x=interpolation_factor * width_integer,
                n_y=interpolation_factor * height_integer,
                order=spline_order,
            )
            distance = np.linspace(0, width, len(intensity))*scaling_factor
            if selected_element.angle % 90 == 0:
                distance += np.min(selected_element.control_points.coords[:, 0, 0])*scaling_factor

        intensity *= 1e-3

        # automatically set correct axis limits
        if resize:
            self.vispy_intensity_plot.diagram.camera.set_range(
                x=(np.min(distance), np.max(distance)),
                y=(np.min(intensity), np.max(intensity)),
            )
        # update the line plot
        self.vispy_intensity_plot.intensity_line.set_data((distance, intensity))
            

    def save_intensity_plot(self):

        file_path = self.main_window.data_handler.save_file_dialog(
            "intensity", "CSV (*.txt *.csv)"
        )
        if not file_path:
            return

        selected_element = self.main_window.vispy_canvas.get_selected_object()
        if isinstance(selected_element, (EditLineVisual, EditRectVisual)):
            image = self.main_window.data_handler.img_data
            interpolation_factor = (
                int(self.ppx_edit.text()) if self.ppx_edit.text() else 1
            )

            if isinstance(selected_element, EditLineVisual):
                length = int(selected_element.length)
                if length <= 0:
                    return
                intensity, eval_coords = selected_element.intensity_profile(
                    image=image, n=interpolation_factor * length
                )

                np.savetxt(
                    file_path,
                    np.column_stack((intensity, eval_coords)),
                    delimiter="\t",
                    header="intensity, pixel_x, pixel_y",
                    fmt="%.2f",
                )

            elif isinstance(selected_element, EditRectVisual):
                width = int(selected_element.width)
                if width <= 0:
                    return
                intensity, eval_coords = selected_element.intensity_profile(
                    image=image,
                    n_x=interpolation_factor * width,
                    n_y=interpolation_factor * int(selected_element.height),
                )

                np.savetxt(
                    file_path,
                    np.column_stack(
                        (intensity, np.reshape(eval_coords, (len(intensity), -1)))
                    ),
                    delimiter="\t",
                    header="intensity, pixel_x, pixel_y, pixel_x, pixel_y, ...",
                    fmt="%.2f",
                )

    def update_colors(self):
        self.vispy_intensity_plot.update_colors()


class IntensityPlot(SceneCanvas):

    CANVAS_SHAPE = (400, 300)
    main_window = None

    def __init__(self, main_window) -> None:

        self.main_window = main_window

        SceneCanvas.__init__(
            self,
            size=self.CANVAS_SHAPE,
        )

        self.unfreeze()
        grid = self.central_widget.add_grid(margin=0)

        # Create a ViewBox and add it to the Grid
        self.diagram = grid.add_view(row=1, col=1, bgcolor="white", camera="panzoom")
        # Create a Line visual and add it to the ViewBox
        self.intensity_line = LinePlot(data=[0, 0], width=1, marker_size=0)
        self.diagram.add(self.intensity_line)

        self.title_label = Label("Intensity Profile", color="black", font_size=8)
        self.title_label.height_max = 20
        grid.add_widget(self.title_label, row=0, col=1)

        # Create AxisWidget objects for the x and y axes and add them to the Grid
        self.x_axis = AxisWidget(
            orientation="bottom",
            axis_font_size=8,
            axis_label_margin=0,
            tick_label_margin=15,
            text_color="black",
        )
        self.xaxis_label = Label("distance (px)", color="black", font_size=8)
        self.xaxis_label.height_max = 20
        grid.add_widget(self.xaxis_label, row=3, col=1)
        self.x_axis.height_max = 30

        self.y_axis = AxisWidget(
            orientation="left",
            axis_font_size=8,
            axis_label_margin=0,
            tick_label_margin=15,
            text_color="black",
        )
        self.y_axis.width_max = 55
        grid.add_widget(self.x_axis, row=2, col=1)
        grid.add_widget(self.y_axis, row=1, col=0)

        # Link the axes to the ViewBox
        self.x_axis.link_view(self.diagram)
        self.y_axis.link_view(self.diagram)

        self.v_line = InfiniteLine(pos=-1, vertical=True, color=(0.7, 0.7, 0.7, 1))
        self.h_line = InfiniteLine(pos=-1, vertical=False, color=(0.8, 0.8, 0.8, 1))

        # Add the lines to the diagram
        self.diagram.add(self.v_line)
        self.diagram.add(self.h_line)

        # add label that displays coordinates of mouse position
        self.mouse_label = Text(
            "",
            color="black",
            font_size=8,
            anchor_x="left",
            anchor_y="top",
            pos=(55, 20),
            parent=self.central_widget,
        )

        self.freeze()

        self.update_colors()

    def update_colors(self):
        if self.main_window.is_dark_mode():
            color = "white"
        else:
            color = "black"

        self.title_label._text_visual.color = color
        self.xaxis_label._text_visual.color = color
        self.intensity_line.set_data(color=color)
        # self.v_line.color = color
        # self.h_line.color = color
        self.mouse_label.color = color
        self.x_axis.axis.text_color = color
        self.x_axis.axis.tick_color = color
        self.y_axis.axis.text_color = color
        self.y_axis.axis.tick_color = color
        bg_color = self.main_window.get_bg_color()
        bg_color = (
            bg_color.red() / 255,
            bg_color.green() / 255,
            bg_color.blue() / 255,
            bg_color.alpha() / 255,
        )
        self.diagram.bgcolor = bg_color
        self.bgcolor = bg_color

    def on_mouse_move(self, event):

        tr = self.scene.node_transform(self.diagram)
        # Get the position of the mouse in data coordinates
        if not (
            (tr.map(event.pos)[:2] > self.diagram.size).any()
            or ((tr.map(event.pos)[:2] < (0, 0)).any())
        ):

            # main use of program
            if event.button == 3:  # mouse wheel button for changing fov
                # enable panning
                self.diagram.camera._viewbox.events.mouse_move.connect(
                    self.diagram.camera.viewbox_mouse_event
                )
                if event.is_dragging:
                    tr = self.diagram.node_transform(self.diagram.scene)
                    dpos = (
                        tr.map(event.last_event.pos)[:2] - tr.map(event.pos)[:2]
                    )  # Calculate the change in position of the mouse
                    self.diagram.camera.pan(
                        dpos
                    )  # Pan the camera based on the mouse movement

            else:
                tr = self.scene.node_transform(self.diagram.scene)
                pos = tr.map(event.pos)
                # disable panning
                self.diagram.camera._viewbox.events.mouse_move.disconnect(
                    self.diagram.camera.viewbox_mouse_event
                )

                # Update the position of the lines
                self.v_line.set_data(pos[0])
                self.h_line.set_data(pos[1])

                # Update the view
                self.update()

                # update coordinates label
                if self.mouse_label.text == "":
                    self.mouse_label.pos = (self.y_axis.width, self.title_label.height)
                self.mouse_label.text = f"{pos[0]:.2f}, {pos[1]:.2f}"
                # Update the position of the label to the
                # upper right corner of the view



# absolute imports
from PySide6.QtOpenGLWidgets import QOpenGLWidget
from PySide6.QtGui import QAction, QIcon, QUndoStack, QPalette
from PySide6.QtWidgets import (
    QApplication,
    QMainWindow,
    QWidget,
    QHBoxLayout,
    QSplitter,
    QMessageBox,
    QToolButton,
    QSizePolicy,
    QStyle,
)
from PySide6.QtCore import Qt
from sys import modules as sys_modules
import numpy as np


# relative imports
from . import measury_path
from .main_ui import MainUI
from .right_ui import RightUI
from .vispy_canvas import VispyCanvas
from .windows import AboutWindow, DataWindow, SettingsWindow, XMLWindow
from .settings import Settings


class MainWindow(QMainWindow):
    """Qt MainWindow of Application.

    This class represents the Main Window of the application and
    contains several subclasses:
        main_ui : Qt UI interface
        vispy_canvas: vispy_implementation into Qt
        data_handler: data structure
    """

    def __init__(self, data_handler, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.data_handler = data_handler

        # settings
        self.settings = Settings(self, "Measury", "Measury")
        self.settings.load_settings()

        # set style
        QApplication.setStyle(self.settings.value("graphics/style"))

        self.setAcceptDrops(True)

        self.vispy_canvas = VispyCanvas(self)
        self.native_vispy_canvas = DropEnabledQOpenGLWidget(
            self.vispy_canvas, parent=self
        )

        self.right_ui = RightUI(self)
        self.main_ui = MainUI(self, parent=self)
        self.vispy_canvas.main_window = self
        self.vispy_canvas.main_ui = self.main_ui

        # create undo stack
        self.undo_stack = QUndoStack(self)

        self.initUI()

    def initUI(self):
        self.setWindowTitle("Measury")
        self.setWindowIcon(QIcon(str(measury_path / "data/tape_measure_128.ico")))
        self.setMinimumSize(700, 700)
        # self.setStyleSheet("background-color: white")

        # Create actions for menu bar
        openAction = QAction(
            self.style().standardIcon(QStyle.StandardPixmap.SP_DialogOpenButton),
            "Open File",
            self,
        )
        openAction.setShortcut("Ctrl+O")
        openAction.setStatusTip("Open document")
        openAction.triggered.connect(lambda: self.main_ui.select_file(file_path=None))

        imageFromClipboardAction = QAction(
            self.style().standardIcon(QStyle.StandardPixmap.SP_FileIcon),
            "Image from Clipboard",
            self,
        )
        imageFromClipboardAction.setShortcut("Ctrl+V")
        imageFromClipboardAction.setStatusTip("Open image from clipboard")
        imageFromClipboardAction.triggered.connect(
            lambda: self.data_handler.open_image_from_clipboard(self.vispy_canvas)
        )

        saveAction = QAction(
            self.style().standardIcon(QStyle.StandardPixmap.SP_DialogSaveButton),
            "Save File",
            self,
        )
        saveAction.setShortcut("Ctrl+S")
        saveAction.setStatusTip("Save")
        saveAction.triggered.connect(self.main_ui.open_save_window)

        seeXMLAction = QAction(
            self.style().standardIcon(QStyle.StandardPixmap.SP_FileDialogInfoView),
            "See XML",
            self,
        )
        seeXMLAction.setShortcut("Ctrl+U")
        seeXMLAction.setStatusTip("See XML of opened file")
        seeXMLAction.triggered.connect(self.open_xml_window)

        # edit actions
        undoAction = QAction(
            self.style().standardIcon(QStyle.StandardPixmap.SP_ArrowBack), "Undo", self
        )
        undoAction.setShortcut("Ctrl+Z")
        undoAction.setStatusTip("Undo last action")
        undoAction.triggered.connect(self.undo_stack.undo)

        redoAction = QAction(
            self.style().standardIcon(QStyle.StandardPixmap.SP_ArrowForward),
            "Redo",
            self,
        )
        redoAction.setShortcut("Ctrl+Y")
        redoAction.setStatusTip("Redo last action")
        redoAction.triggered.connect(self.undo_stack.redo)

        # view
        centerAction = QAction(
            self.style().standardIcon(QStyle.StandardPixmap.SP_DesktopIcon),
            "Center Image",
            self,
        )
        centerAction.setShortcut("Ctrl+P")
        centerAction.setStatusTip("Center Image")
        centerAction.triggered.connect(self.vispy_canvas.center_image)

        hideAction = QAction(
            QIcon(str(measury_path / "data/ghost_icon.png")), "Hide Objects", self
        )
        hideAction.setShortcut("Ctrl+H")
        hideAction.setStatusTip("Hide Objects")
        hideAction.triggered.connect(self.vispy_canvas.hide_all_objects_w_undo)

        showAction = QAction(
            QIcon(str(measury_path / "data/show_icon.png")), "Show Objects", self
        )
        showAction.setShortcut("Ctrl+Shift+H")
        showAction.setStatusTip("Show Objects")
        showAction.triggered.connect(self.vispy_canvas.show_all_objects_w_undo)

        # Settings
        settingsAction = QAction(
            QIcon(str(measury_path / "data/gear_icon.png")), "Settings", self
        )
        settingsAction.setShortcut("Ctrl+I")
        settingsAction.setStatusTip("Open settings page")
        settingsAction.triggered.connect(self.open_settings_page)

        resetSettingsAction = QAction(
            self.style().standardIcon(QStyle.StandardPixmap.SP_DialogResetButton),
            "Reset Settings",
            self,
        )
        resetSettingsAction.setStatusTip("Clear Settings")
        resetSettingsAction.triggered.connect(self.settings.clear)

        # About
        aboutAction = QAction(
            self.style().standardIcon(QStyle.StandardPixmap.SP_MessageBoxInformation),
            "About",
            self,
        )
        aboutAction.setStatusTip("Show information about Measury")
        aboutAction.triggered.connect(self.open_about_page)

        # Data Overview

        measurementAction = QAction(
            self.style().standardIcon(QStyle.StandardPixmap.SP_FileDialogDetailedView),
            "Data Overview",
            self,
        )
        measurementAction.setShortcut("Ctrl+D")
        measurementAction.setStatusTip("Show an overview of all measured data")
        measurementAction.triggered.connect(self.open_data_page)

        # export coordinates action

        exportCoordsAction = QAction(
            self.style().standardIcon(QStyle.StandardPixmap.SP_FileDialogListView),
            "Export Coordinates",
            self,
        )
        exportCoordsAction.setShortcut("Ctrl+Shift+E")
        exportCoordsAction.setStatusTip("Export coordinates of selected object")
        exportCoordsAction.triggered.connect(self.export_object_coords)


        delete_all_objects_action = QAction(
            self.style().standardIcon(QStyle.StandardPixmap.SP_DialogDiscardButton),
            "Delete all Objects",
            self,
        )
        delete_all_objects_action.setStatusTip("Delete all objects")
        delete_all_objects_action.triggered.connect(
            self.data_handler.delete_all_objects_w_undo
        )

        # Create menu bar
        menuBar = self.menuBar()
        fileMenu = menuBar.addMenu("&File")
        fileMenu.addAction(openAction)
        fileMenu.addAction(imageFromClipboardAction)
        fileMenu.addAction(saveAction)
        fileMenu.addAction(seeXMLAction)

        editMenu = menuBar.addMenu("&Edit")
        editMenu.addAction(delete_all_objects_action)
        editMenu.addAction(undoAction)
        editMenu.addAction(redoAction)
        settingsMenu = menuBar.addMenu("&Settings")
        settingsMenu.addAction(settingsAction)
        settingsMenu.addAction(resetSettingsAction)
        measurementsMenu = menuBar.addMenu("&Measurements")
        measurementsMenu.addAction(measurementAction)
        measurementsMenu.addAction(exportCoordsAction)
        viewMenu = menuBar.addMenu("&View")
        viewMenu.addAction(centerAction)
        viewMenu.addAction(hideAction)
        viewMenu.addAction(showAction)
        miscMenu = menuBar.addMenu("&Misc")
        miscMenu.addAction(aboutAction)

        central_widget = QWidget()
        main_layout = QHBoxLayout()

        # splitter
        self.splitter = QSplitter(Qt.Orientation.Horizontal)

        # controls
        self.splitter.addWidget(self.main_ui)
        # vispy canvas
        self.splitter.addWidget(self.native_vispy_canvas)
        # right ui
        self.splitter.addWidget(self.right_ui)

        # The slim bar (tool button)
        self.open_right_ui_button = QToolButton()
        self.open_right_ui_button.setText("<")
        self.open_right_ui_button.clicked.connect(self.right_ui.show_ui)

        self.open_right_ui_button.setSizePolicy(
            QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding
        )
        # Hide the right panel by default
        self.right_ui.hide()

        main_layout.addWidget(self.splitter)
        main_layout.addWidget(self.open_right_ui_button)

        central_widget.setLayout(main_layout)
        self.setCentralWidget(central_widget)

    def open_about_page(self):
        self.about_window = AboutWindow(parent=self)
        self.about_window.show()
        self.about_window.activateWindow()

    def open_settings_page(self):
        self.get_bg_color()
        self.about_window = SettingsWindow(parent=self)
        self.about_window.show()
        self.about_window.activateWindow()

    def open_data_page(self):
        if self.data_handler.drawing_data:
            self.data_window = DataWindow(parent=self)
            self.data_window.show()
            self.data_window.activateWindow()

    def closeEvent(self, event):
        QApplication.closeAllWindows()

    def raise_error(self, error):
        self.data_handler.logger.error(error.replace("\n", " "))
        if "pytest" in sys_modules:
            raise Exception(error)
        else:
            QMessageBox.critical(self, "Error", str(error))

    def reset_undo_stack(self):
        self.undo_stack.clear()

    def keyPressEvent(self, event):
        if event.key() == Qt.Key.Key_Delete:
            if not self.vispy_canvas.start_state:
                if self.vispy_canvas.selected_object is not None:
                    # TODO: Setting or onetime warning for delete
                    # reply = QMessageBox.warning(self, "Warning",
                    #     "Do you want to delete the selected object?",
                    #     QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                    #     QMessageBox.StandardButton.Yes)

                    # if reply == QMessageBox.StandardButton.Yes:
                    self.vispy_canvas.delete_object_w_undo()

    def open_xml_window(self):
        self.data_window = XMLWindow(parent=self)
        self.data_window.show()

    def update_style(self):
        self.data_handler.logger.info(
            "Updating style to "
            + self.settings.value("graphics/style")
            + f" dark_mode = {self.is_dark_mode()}"
        )
        QApplication.setStyle(self.settings.value("graphics/style"))

    def is_dark_mode(self):
        app = (
            QApplication.instance()
        )  # Ensures it works with the current QApplication instance
        if not app:  # If the application does not exist, create a new instance
            app = QApplication([])
        return app.palette().color(QPalette.ColorRole.Window).lightness() < 128

    def get_bg_color(self):
        QApplication.processEvents()
        color = self.palette().color(QPalette.ColorRole.Window)
        return color
    
    def export_object_coords(self):
        
        object = self.vispy_canvas.get_selected_object()
        if object is None:
            self.raise_error("No object selected")
            return
        else:
            coords = object.control_points.get_coords()
            # open file dialog
            file_path = self.data_handler.save_file_dialog(
                file_name="coords.txt",
                extensions="Measurement File (*.txt);;All Files (*)",
            )
            if file_path is None:
                return

            with open(file_path, "wb") as save_file:
                # write to text file
                self.data_handler.logger.info(f"Saving coordinates to {file_path}")
                if self.main_ui.scaling_factor is not None:
                    coords = coords * self.main_ui.scaling_factor
                    header = f"Coordinates in {self.main_ui.units_dd.currentText()}, scaled by {self.main_ui.scaling_factor} {self.main_ui.units_dd.currentText()}/px"
                else:
                    header = f"Coordinates in px"
                np.savetxt(save_file, coords, delimiter=",", header=header)


class DropEnabledQOpenGLWidget(QOpenGLWidget):
    def __init__(self, vispy_canvas, parent=None):
        super().__init__(parent)
        self.setAcceptDrops(True)
        self.vispy_canvas = vispy_canvas
        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)  # Remove margins
        layout.addWidget(vispy_canvas.native)

    def dragEnterEvent(self, event):
        if len(event.mimeData().urls()) == 1:
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        if len(event.mimeData().urls()) == 1:
            img_path = event.mimeData().urls()[0].toLocalFile()
            self.vispy_canvas.data_handler.open_file(img_path)



# absolute imports
from vispy.app import use_app
import sys
import warnings

# relative imports
from .main_window import MainWindow
from .data_handler import DataHandler


class App:

    def __init__(self, file_path=None, logger=None):

        self.vispy_app = use_app("pyside6")
        self.vispy_app.create()
        self.data_handler = DataHandler(logger)
        self.main_window = MainWindow(self.data_handler)
        self.data_handler.main_window = self.main_window
        self.data_handler.open_file(
            file_path=file_path
        )

    def run(self, run_vispy=True):
        """
        Run the application.

        Args:
            run_vispy (bool, optional): Whether to run the VisPy application.
            Defaults to True.
        """
        sys.excepthook = self.exception_hook
        warnings.showwarning = self.warning_handler

        self.main_window.show()
        # update axis because when creating the qt application the
        # transformations change implicitely
        self.main_window.vispy_canvas.center_image()
        if run_vispy:
            self.vispy_app.run()

    def exception_hook(self, exception_type, exception_value: Exception, traceback):
        """
        Exception hook for the application.

        Args:
            exception_type (type): Type of the exception.
            exception_value (Exception): The exception.
            traceback (traceback): The traceback of the exception.
        """
        # Print the error message to stderr
        print(f"Unhandled exception: {exception_value}")

        # Display an error message box
        self.main_window.raise_error(exception_value)
        sys.__excepthook__(exception_type, exception_value, traceback)
        
    def warning_handler(self, message, category, filename, lineno, file=None, line=None):
        """
        Custom warning handler.

        Args:
            message (Warning): The warning message.
            category (Warning): The category of the warning.
            filename (str): The file name where the warning occurred.
            lineno (int): The line number where the warning occurred.
            file (file object, optional): The file object to write the warning to.
            line (str, optional): The line of code where the warning occurred.
        """
        warning_message = f"{filename}:{lineno}: {category.__name__}: {message}"
        print(warning_message)

    def close(self):
        """
        Closes the application by quitting the Vispy app and
        closing the main window.
        """
        self.vispy_app.quit()
        self.main_window.close()


def run(file_path=None, logger=None):
    if len(sys.argv) > 1:
        app = App(file_path=sys.argv[1], logger=logger)
    else:
        app = App(file_path=file_path, logger=logger)

    app.run()



from pathlib import Path

try:
    from ._version import version as __version__
    from ._version import version_tuple
except ImportError:
    __version__ = "unknown version"
    version_tuple = (0, 0, "unknown version")

# for compiling the program
measury_path = Path(__file__).parent.resolve()

from .app import run


from PySide6.QtGui import QColor
from PySide6.QtCore import QSettings

DEFAULT_SETTINGS = {
    "graphics/object_color": QColor(255, 0, 0, 26),
    "graphics/object_border_color": QColor(255, 0, 0, 127),
    "graphics/image_rendering": "nearest",
    "graphics/scale_bar_color": QColor(255, 0, 0, 255),
    "graphics/style": "Fusion",
    "ui/microscope": "Generic_Microscope",
    "ui/show_both_scaling": False,
    "misc/file_extensions": [".msry", ".measury"],
}


class Settings(QSettings):

    def __init__(self, parent, organization: str, application: str):
        super().__init__(organization, application)
        self.parent = parent

    def load_defaults(self):
        """Manage loading all the settings."""
        self.parent.data_handler.logger.info("Loading default settings")
        for key, value in DEFAULT_SETTINGS.items():
            self.setValue(key, value)
        self.sync()

    def load_settings(self):
        """Load all the settings."""
        self.parent.data_handler.logger.info("Loading settings")
        for key in DEFAULT_SETTINGS.keys():
            if self.value(key) is None:
                # No settings exist, load the default settings
                self.load_defaults()

    def save(self, key, value):
        """Manage saving all the settings."""
        self.parent.data_handler.logger.info(f"Saving setting: {key}")
        self.setValue(key, value)
        self.sync()

    @property
    def is_default(self):
        """Test if the settings are the default values."""
        return self.equals_settings(DEFAULT_SETTINGS)

    def equals_settings(self, settings: dict):
        """Compare the settings to the default settings."""
        # check if the keys are the same
        if set(self.allKeys()) != set(settings.keys()):
            self.parent.data_handler.logger.info(
                f"Settings not equal: {set(self.allKeys())} {set(settings.keys())}"
            )
            return False
        # check if the values are the same
        for key in settings.keys():
            # for lists we need to compare the sets
            if isinstance(settings.get(key), list):
                if set(self.value(key)) != set(settings.get(key)):
                    self.parent.data_handler.logger.info(
                        f"Settings not equal: {set(self.allKeys())} {set(settings.keys())}"
                    )
                    return False
            # for boolean we need to set the type
            elif isinstance(settings.get(key), bool):
                if self.value(key, type=bool) != settings.get(key):
                    self.parent.data_handler.logger.info(
                        f"Setting not equal: {key} {self.value(key)} {settings.get(key)}"
                    )
                    return False
            # otherwise simply compare the values
            elif self.value(key) != settings.get(key):
                self.parent.data_handler.logger.info(
                    f"Setting not equal: {key} {self.value(key)} {settings.get(key)}"
                )
                return False
        return True



# absolute imports
import numpy as np
import cv2
from vispy.scene import SceneCanvas, visuals, AxisWidget, Label
from vispy.visuals.transforms import linear
from PySide6.QtWidgets import QInputDialog
from PySide6.QtGui import QUndoCommand

# relative imports
from .drawable_objects import (
    EditEllipseVisual,
    EditRectVisual,
    ControlPoints,
    EditLineVisual,
    LineControlPoints,
    EditPolygonVisual,
)

class VispyCanvas(SceneCanvas):
    """Canvas for displaying the vispy instance"""

    CANVAS_SHAPE = (800, 600)
    main_ui = None
    main_window = None
    scale_bar_params = (None, True, None, None) # seed point, relative, threshold, direction
    text_color = "black"
    current_move = None
    origin = np.zeros(2)

    def __init__(self, main_window):

        self.data_handler = main_window.data_handler
        self.main_window = main_window

        SceneCanvas.__init__(
            self, size=self.CANVAS_SHAPE, keys=dict(delete=self.delete_object_w_undo)
        )
        self.unfreeze()
        # get background color of qt session
        self.update_background_color()

        if self.main_window.is_dark_mode():
            self.text_color = "white"

        self.grid = self.central_widget.add_grid(margin=0)

        self.view = self.grid.add_view(row=1, col=1, bgcolor="black")

        self.image = visuals.Image(
            data=None,
            texture_format="auto",
            interpolation=self.main_window.settings.value("graphics/image_rendering"),
            cmap="viridis",
            parent=self.view.scene,
        )

        # title
        self.title_label = Label(
            "Microscope-Image", color=self.text_color, font_size=12
        )
        self.title_label.height_max = 40
        self.grid.add_widget(self.title_label, row=0, col=0, col_span=3)

        # y axis
        self.yaxis = AxisWidget(
            orientation="left",
            axis_font_size=12,
            axis_label_margin=0,
            tick_label_margin=15,
            text_color=self.text_color,
        )
        self.yaxis.width_max = 80
        self.grid.add_widget(self.yaxis, row=1, col=0)

        # x axis
        self.xaxis = AxisWidget(
            orientation="bottom",
            axis_font_size=12,
            axis_label_margin=0,
            tick_label_margin=15,
            text_color=self.text_color,
        )
        self.xaxis.height_max = 50
        self.grid.add_widget(self.xaxis, row=2, col=1)

        self.update_axis_color()

        # padding right
        right_padding = self.grid.add_widget(row=1, col=2, row_span=2)
        right_padding.width_max = 5

        self.load_image_label = Label(
            "Click here to select\nan image", color="white", font_size=16
        )
        self.grid.add_widget(self.load_image_label, row=1, col=1)

        # camera
        self.view.camera.set_range(x=(0, 1), y=(0, 1), margin=0)

        self.start_state = True
        self.update_image()

        # for manipulating shapes
        self.selected_object = None
        self.selected_point = None

        self.freeze()

    def update_image(self):
        """Update the image in the vispy canvas and also the file path shown above the image"""

        # if there is no image loaded, do nothing
        if self.data_handler.file_path is not None:
            try:
                # opencv reads images in BGR format,
                # so we need to convert it to RGB
                self.data_handler.logger.debug(
                    f"update data_handler.file_path = {self.data_handler.file_path}"
                )
                BGR_img = cv2.imdecode(
                    np.frombuffer(self.data_handler.img_byte_stream, dtype=np.uint8),
                    cv2.IMREAD_UNCHANGED,
                )
                self.data_handler.img_data = cv2.cvtColor(BGR_img, cv2.COLOR_BGR2RGB)

                self.draw_image()

                self.view.camera = "panzoom"
                self.center_image()
                self.view.camera.aspect = 1
                self.view.camera.flip = (0, 1, 0)

                self.xaxis.link_view(self.view)
                self.yaxis.link_view(self.view)

                self.title_label.text = self.data_handler.file_path.name

            except Exception as error:
                # handle the exception
                if self.main_window is not None:
                    self.main_window.raise_error(f"Image could not be loaded: {error}")
                    return

            self.start_state = False
            self.remove_load_text()

    def draw_image(self, img_data=None):
        if img_data is None:
            img_data = self.data_handler.img_data
        self.data_handler.logger.debug("set vispy image data")
        self.image.set_data(img_data)
        self.image.interpolation = self.main_window.settings.value(
            "graphics/image_rendering"
        )

    def center_image(self):
        try:
            if self.data_handler.img_data is not None:
                self.view.camera.set_range(
                    x=(0-self.origin[0], self.data_handler.img_data.shape[1]-self.origin[0]),
                    y=(0-self.origin[1], self.data_handler.img_data.shape[0]-self.origin[1]),
                    margin=0,
                )
        except Exception as error:
            self.main_window.raise_error(f"Image could not be centered: {error}")

    def add_load_text(self):
        # load image label
        self.load_image_label.visible = True

    def remove_load_text(self):
        # self.load_image_label.text = ""
        self.load_image_label.visible = False

    def on_mouse_release(self, event):
        # transform so that coordinates start at 0 in self.view window
        tr = self.scene.node_transform(self.view)
        # only activate when over self.view by looking
        # if coordinates < 0 or > size of self.view
        if not (
            (tr.map(event.pos)[:2] > self.view.size).any()
            or (tr.map(event.pos)[:2] < (0, 0)).any()
        ):

            if self.start_state:
                # when no image is selected open file opening
                #  sequence by clicking
                self.main_ui.select_file()
            elif event.button == 1:
                match self.main_ui.tools_buttons.checkedButton().text(): 
                    case "Line" | "Ellipse" | "Rectangle" | "Angle" | "Multi-Line" | "Polygon" | "Edit":
                        
                        # multi-line is not finished yet
                        if isinstance(self.selected_object, LineControlPoints):
                            if self.selected_object.continue_adding_points:
                                tr = self.scene.node_transform(self.selected_object)
                                pos = tr.map(event.pos)
                                self.add_point_w_undo(self.selected_object, pos[:2])
                            elif self.current_move.check_movement():
                                self.main_window.undo_stack.push(self.current_move)
                                
                                
                        # redo move object command to save current location
                        elif self.current_move is not None:
                            # if object was moved, put it on undo stack
                            if self.current_move.check_movement():
                                self.main_window.undo_stack.push(self.current_move)

        # if click is above box
        elif (
            tr.map(event.pos)[1] < 0 # above box
            and event.pos[1] > 0 # but still in vispy canvas
            and tr.map(event.pos)[0] > 0 # right of start of scene
            and tr.map(event.pos)[0] < self.view.size[0] # left of end of scene
        ):
            if not self.data_handler.file_path:
                self.main_ui.select_file()
            elif self.data_handler.file_path.name != "clipboard":
                self.data_handler.open_file_location(self.data_handler.file_path)
                        

    def on_mouse_press(self, event):
        # transform so that coordinates start at 0 in self.view window
        tr = self.scene.node_transform(self.view)
        # only activate when over self.view by looking
        #  if coordinates < 0 or > size of self.view
        if not (
            (tr.map(event.pos)[:2] > self.view.size).any()
            or (tr.map(event.pos)[:2] < (0, 0)).any()
        ):

            # if in start state, dont do anything as we are looking
            # at mouse_released for that
            if self.start_state:
                return

            # main use of program
            if event.button == 3:  # mouse wheel button for changing fov
                # enable panning
                self.view.camera._viewbox.events.mouse_move.connect(
                    self.view.camera.viewbox_mouse_event
                )

            else:
                match self.main_ui.tools_buttons.checkedButton().text():

                    case "Move":
                        # enable panning
                        self.view.camera._viewbox.events.mouse_move.connect(
                            self.view.camera.viewbox_mouse_event
                        )

                        # QApplication.setOverrideCursor(QCursor(Qt.CursorShape.SizeAllCursor))

                        # unselect object
                        self.unselect()

                    case "Set Origin":
                        # disable panning
                        self.view.camera._viewbox.events.mouse_move.disconnect(
                            self.view.camera.viewbox_mouse_event
                        )

                        tr = self.scene.node_transform(self.view.scene)
                        pos = tr.map(event.pos)[:2]

                        # unselect object
                        self.unselect()

                        self.set_origin_w_undo(pos+self.origin)

                    case "Select":
                        # disable panning
                        self.view.camera._viewbox.events.mouse_move.disconnect(
                            self.view.camera.viewbox_mouse_event
                        )

                        tr = self.scene.node_transform(self.view.scene)
                        pos = tr.map(event.pos)
                        self.view.interactive = False
                        selected = self.visual_at(event.pos)
                        self.view.interactive = True

                        if event.button == 1:

                            if selected is not None:
                                # unselect object to create room of new one
                                self.unselect()

                                self.selected_object = selected.parent

                                # update transform to selected object
                                tr = self.scene.node_transform(self.selected_object)
                                pos = tr.map(event.pos)

                                self.selected_object.select(True, obj=selected)
                                self.selected_object.start_move(pos)
                        
                                # update ui to display properties of selected object
                                self.selection_update()

                    case "Line" | "Ellipse" | "Rectangle" | "Angle" | "Multi-Line" | "Polygon" | "Edit":
                        # disable panning
                        self.view.camera._viewbox.events.mouse_move.disconnect(
                            self.view.camera.viewbox_mouse_event
                        )

                        tr = self.scene.node_transform(self.view.scene)
                        pos = tr.map(event.pos)
                        self.view.interactive = False
                        selected = self.visual_at(event.pos)
                        self.view.interactive = True
                        # unselect object to create room of new one
                        self.unselect()
                                

                        if event.button == 1:
                            # QApplication.setOverrideCursor(QCursor(Qt.CursorShape.SizeAllCursor))

                            # edit object
                            if selected is not None:
                                self.selected_object = selected.parent
                                # update transform to selected object
                                tr = self.scene.node_transform(self.selected_object)
                                pos = tr.map(event.pos)

                                self.selected_object.select(True, obj=selected)
                                self.selected_object.start_move(pos)
                                # self.move_object_w_undo(self.selected_object)
                                
                                if not isinstance(self.selected_object, LineControlPoints):    
                                    self.current_move = MoveObjectCommand(self, self.selected_object)
                                    self.data_handler.logger.debug("start moving object")
                                
                                elif not self.selected_object.continue_adding_points:
                                    self.current_move = MoveObjectCommand(self, self.selected_object)
                                    self.data_handler.logger.debug("start moving object")

                                # update ui to display properties of selected object
                                self.selection_update()
                            else:
                                self.current_move = None
                            
                            # create new object:
                            if self.selected_object is None:

                                match self.main_ui.tools_buttons.checkedButton().text():
                                    case "Line":
                                        new_object = EditLineVisual(
                                            parent=self.view.scene,
                                            settings=self.main_window.settings,
                                            num_points=2,
                                        )
                                    case "Ellipse":
                                        new_object = EditEllipseVisual(
                                            parent=self.view.scene,
                                            settings=self.main_window.settings,
                                        )
                                    case "Rectangle":
                                        new_object = EditRectVisual(
                                            parent=self.view.scene,
                                            settings=self.main_window.settings,
                                        )
                                    case "angle":
                                        new_object = EditLineVisual(
                                            settings=self.main_window.settings,
                                            parent=self.view.scene,
                                            num_points=3,
                                        )
                                    case "Multi-Line":
                                        new_object = EditLineVisual(
                                            settings=self.main_window.settings,
                                            parent=self.view.scene,
                                            num_points=0,
                                        )
                                    case "Polygon":
                                        new_object = EditPolygonVisual(
                                            settings=self.main_window.settings,
                                            parent=self.view.scene)
                                    case "Edit":
                                        return
                                # dont show object before it is added to drawing_data
                                new_object.select(False)
                                if self.check_creation_allowed(new_object):

                                    self.create_new_object_w_undo(
                                        new_object, pos=pos, selected=True
                                    )

                                    # update ui where data is shown
                                    self.selection_update(object=new_object)
                                    # if linecontrolpoint to get automatic line creation without draggin
                                    if isinstance(self.selected_object, LineControlPoints):
                                        last_cp = self.selected_object.control_points[-1]
                                        self.selected_object.select(True, last_cp)
                                else:
                                    new_object.delete()

                                

                        # delete object
                        if event.button == 2:  # right button deletes object
                            # not self.selected_object because we want to
                            # delete it on hover too
                            if selected is not None:
                                self.delete_object_w_undo(object=selected.parent)

                                # Needs change
                                # self.main_ui.update_save_window()

                    case "Identify Scaling":
                        # disable panning
                        self.view.camera._viewbox.events.mouse_move.disconnect(
                            self.view.camera.viewbox_mouse_event
                        )

                        if event.button == 1:

                            # transform to get image pixel coordinates
                            tr_image = self.scene.node_transform(self.image)
                            mouse_image_coords = tr_image.map(event.pos)[:2]
                            # mouse_click_coordinates
                            m_i_x, m_i_y = np.floor(mouse_image_coords).astype(int)

                            # get width of scaling bar and show it in image
                            self.find_scale_bar_width_w_undo(
                                (m_i_x, m_i_y), 
                                relative=False, 
                                threshold=self.main_ui.get_threshold(),
                                direction=self.main_ui.scaling_direction_dd.currentText()
                            )

                        # right click to delete scaling identification
                        if event.button == 2:
                            self.draw_image(self.data_handler.img_data)
                            self.main_ui.pixel_edit.setText(None)

    def create_new_object(
        self, new_object, pos=None, selected=False, structure_name=None
    ):

        self.data_handler.logger.info("Creating new Object")

        if pos is not None:
            new_object.set_center(pos[0:2])

        new_object.select(selected)
        if selected:
            new_object.select_creation_controlpoint()
            self.selected_object = new_object.control_points

        # save object in data_handler.drawing_data
        if structure_name is None:
            structure_name = self.main_ui.structure_dd.currentText()

        structure_name = self.data_handler.save_object(structure_name, new_object)

        if self.main_ui.structure_dd.currentText() == "":
            self.main_ui.structure_dd.setCurrentText(structure_name)

        self.main_ui.update_object_list()

        return structure_name

    def create_new_object_w_undo(
        self, new_object, pos=None, selected=False, structure_name=None
    ):
        command = CreateObjectCommand(self, new_object, pos, selected, structure_name)
        self.main_window.undo_stack.push(command)

    def update_object_colors(self):
        for name in self.data_handler.drawing_data.keys():
            for obj in self.data_handler.drawing_data[name]:
                color = self.main_window.settings.value(
                    "graphics/object_color"
                ).getRgb()
                color = tuple([value / 255 for value in color])
                border_color = self.main_window.settings.value(
                    "graphics/object_border_color"
                ).getRgb()
                border_color = tuple([value / 255 for value in border_color])

                if isinstance(obj, (EditLineVisual)):
                    obj.line_color = border_color
                    obj.update_from_controlpoints()
                else:
                    obj.update_colors(color=color, border_color=border_color)
        self.scene.update()

    def update_background_color(self, color=None):
        if color is None:
            color = self.main_window.get_bg_color()
        # if self.main_window.is_dark_mode():
        #     color = "black"
        # else:
        #     color = (240/255, 240/255, 240/255, 1)

        # Convert QColor to RGBA tuple with values in range 0-1
        self.bgcolor = (
            color.red() / 255,
            color.green() / 255,
            color.blue() / 255,
            color.alpha() / 255,
        )
        self.update()

    def update_axis_color(self):

        if self.main_window.is_dark_mode():
            color = "white"
        else:
            color = "black"
        # Convert QColor to RGBA tuple with values in range 0-1
        # self.title_label.color = color
        self.xaxis.axis.text_color = color
        # self.xaxis.axis.axis_color = color
        self.xaxis.axis.tick_color = color
        self.yaxis.axis.text_color = color
        # self.yaxis.axis.axis_color = color
        self.yaxis.axis.tick_color = color

        # title color
        self.title_label._text_visual.color = color

    def update_colors(self):
        self.update_background_color()
        self.update_axis_color()
        self.update_object_colors()

    def check_creation_allowed(self, new_object, structure_name=None):
        # check if object is allowed to be created
        if structure_name is None:
            structure_name = self.main_ui.structure_dd.currentText()

        if structure_name in self.data_handler.drawing_data.keys():
            structure_type = type(self.data_handler.drawing_data[structure_name][0])
        else:
            return True

        if isinstance(new_object, (structure_type)):
            return True
        else:
            # raise prompt to change structure name
            text, ok = QInputDialog.getText(
                self.main_ui,
                "Warning",
                f"This structure is of a different type: {structure_type}.\n"
                "Please change the structure name to create this object.",
                text=self.data_handler.generate_output_name(),
            )
            if ok:
                # After the user closes the QInputDialog,
                # you can get the text from the QLineEdit
                self.main_ui.structure_dd.setCurrentText(text)
            return False

    def find_scale_bar_width(self, seed_points, relative=True, threshold=None, direction=None):
        # get width of scaling bar by floodFilling an area of similar pixels.
        # The start point needs to be given

        if self.start_state:
            return
        if threshold is None:
            threshold = self.main_ui.DEFAULT_THRESHOLD
        if direction is None:
            direction = 'horizontal'
        
        self.scale_bar_params = (seed_points, relative, threshold, direction)
        self.main_ui.set_threshold(threshold)
        self.main_ui.scaling_direction_dd.setCurrentText(direction)

        if seed_points is None:
            self.draw_image()
            scale_px = ""
        else:
            # create copy of image which can be modified
            img_data_modified = self.data_handler.img_data.copy()
            if relative:
                seed_point_x = round(seed_points[0] * img_data_modified.shape[1])
                seed_point_y = round(seed_points[1] * img_data_modified.shape[0])
            else:
                seed_point_x, seed_point_y = seed_points
            self.data_handler.logger.debug(
                f"seed_point_x = {seed_point_x}, seed_point_y = {seed_point_y}"
            )

            cv2.floodFill(
                img_data_modified,
                None,
                (seed_point_x, seed_point_y),
                newVal=self.main_window.settings.value(
                    "graphics/scale_bar_color"
                ).getRgb(),
                loDiff=[threshold] * 3,
                upDiff=[threshold] * 3,
            )

            self.draw_image(img_data=img_data_modified)

            non_zero_indices = np.nonzero(
                np.sum(self.data_handler.img_data - img_data_modified, axis=2)
            )

            # plus one to account for the start pixel
            if direction == "horizontal":
                scale_px = np.max(non_zero_indices[1]) - np.min(non_zero_indices[1]) + 1
            else:
                scale_px = np.max(non_zero_indices[0]) - np.min(non_zero_indices[0]) + 1

        self.main_ui.pixel_edit.setText(str(scale_px))
        self.scene.update()

    def find_scale_bar_width_w_undo(self, seed_point_percentage, relative=True, threshold=None, direction=None):
        command = FindScalingBarWidthCommand(
            self, seed_point_percentage, relative, threshold, direction
        )
        self.main_window.undo_stack.push(command)

    def on_key_press(self, event):

        if self.start_state:
            return

        if isinstance(self.selected_object, LineControlPoints):
            if event.key.name == "Escape":
                if self.selected_object.continue_adding_points:
                    # self.selected_object.remove_point()
                    self.remove_point_w_undo()
                    self.selected_object.continue_adding_points = False
                else:
                    self.remove_point_w_undo()
            if event.key.name == "A":
                if (self.selected_object.num_points == 0 
                or len(self.selected_object.coords) < self.selected_object.num_points):

                    self.selected_object.continue_adding_points = True
                    
                    # get mouse position in image coordinates
                    index = self.selected_object.get_selected_index()
                    pos = self.selected_object.coords[index]
                    # undo gets called, when the released event is called
                    self.add_point_w_undo(self.selected_object, pos)


    def on_mouse_move(self, event):
        
        # transform so that coordinates start at 0 in self.view window
        tr = self.scene.node_transform(self.view)
        # only activate when over self.view by looking
        # if coordinates < 0 or > size of self.view
        if not (
            (tr.map(event.pos)[:2] > self.view.size).any()
            or (tr.map(event.pos)[:2] < (0, 0)).any()
        ):

            if self.start_state:
                return
            match event.button:
                case 1:
                    # Check if the left mouse button is being dragged
                    if not event.is_dragging:
                        return

                    # scale objects while dragging
                    if self.selected_object is not None:
                        modifiers = [key.name for key in event.modifiers]
                        tr = self.scene.node_transform(self.selected_object)
                        pos = tr.map(event.pos)

                        if self.main_ui.tools_buttons.checkedButton().text() in (
                            "Line", "Ellipse", "Rectangle", "Angle", "Edit", "Polygon", "Multi-Line"
                        ):

                            if "Shift" in modifiers and not isinstance(
                                self.selected_object,
                                (LineControlPoints, EditLineVisual),
                            ):

                                # calculate angle of current position
                                x, y = pos[0:2] - self.selected_object.center
                                angle = np.arctan2(-y, x)

                                # Check if "Control" is pressed for snapping
                                if "Control" in modifiers:
                                    # Calculate the nearest multiple of 45° (np.pi / 4 radians)
                                    nearest_multiple = round(angle / (np.pi / 4)) * (np.pi / 4)
                                    # Snap to the nearest multiple if the angle is close to it
                                    if abs(angle - nearest_multiple) < np.deg2rad(10):
                                        angle = nearest_multiple + self.selected_object.drag_reference_angle

                                self.selected_object.rotate(angle)

                            else:

                                self.selected_object.move(pos[0:2], modifiers=modifiers)

                            # update ui to display properties of selected object
                            self.selection_update()
                            self.main_window.right_ui.update_intensity_plot()

                case 3:
                    if event.is_dragging:
                        tr = self.view.node_transform(self.view.scene)
                        # Calculate the change in position of the mouse
                        dpos = tr.map(event.last_event.pos)[:2] - tr.map(event.pos)[:2]
                        # Pan the camera based on the mouse movement
                        self.view.camera.pan(dpos)
                
                case _:
                    # if it is a line that is still being drawn
                    if isinstance(self.selected_object, LineControlPoints):

                        if self.selected_object.continue_adding_points:
                            modifiers = [key.name for key in event.modifiers]
                            tr = self.scene.node_transform(self.selected_object)
                            pos = tr.map(event.pos)

                            self.selected_object.move(pos[:2], modifiers=modifiers)
                            self.selection_update()
                            self.main_window.right_ui.update_intensity_plot()
                        
    def hide_arrows(self):
        for name in self.data_handler.drawing_data.keys():
            for obj in self.data_handler.drawing_data[name]:
                if isinstance(obj, EditRectVisual):
                    obj.hide_arrow()
        self.scene.update()
        
    def show_arrows(self):
        for name in self.data_handler.drawing_data.keys():
            for obj in self.data_handler.drawing_data[name]:
                if isinstance(obj, EditRectVisual):
                    obj.show_arrow()
        self.scene.update()

    def move_object_w_undo(self, object):
        command = MoveObjectCommand(self, object)
        self.main_window.undo_stack.push(command)

    def remove_point_w_undo(self):
        # we dont want to undo the last 2 points
        if len(self.selected_object.coords) < 3:
            self.selected_object.remove_point()
        else:
            command = RemovePointCommand(self, self.selected_object)
            self.main_window.undo_stack.push(command)

    def add_point_w_undo(self, object, point):
        # we dont want to remember creating the first 2 (3 for angle) points 
        # as this is basically the creation of the object
        if (len(self.selected_object.coords) < 2 
        or (self.selected_object.num_points != 0 
            and len(self.selected_object.coords) < self.selected_object.num_points+1)):
            self.selected_object.add_point(point)
        else:
            command = AddPointCommand(self, object, point)
            self.main_window.undo_stack.push(command)

    def selection_update(self, object=None):

        if object is None:
            object = self.get_selected_object()
        if self.selected_object is None:
            self.main_ui.clear_object_table()
            return

        self.main_window.right_ui.update_intensity_plot()

        found_object = self.data_handler.find_object(object)
        if found_object:
            structure, index = found_object
        else:
            return
        self.main_ui.add_to_structure_dd(structure)
        self.main_ui.update_object_list()
        self.main_ui.object_list.item(index).setSelected(True)

        self.main_ui.update_object_table(object)

    def update_object_property_w_undo(self, obj, prop, value, scaling_factor=None, old_value=None):
        # only push onto undo stack if property is modifiable
        if prop in obj.get_modifiable_properties():
            command = UpdateObjectCommand(self, obj, prop, value, scaling_factor, old_value)
            self.main_window.undo_stack.push(command)
        # otherwise just redraw old value
        else:
            self.selection_update()

    def update_object_property(self, obj, prop, value, scaling_factor=None):
        if obj is None:
            obj = self.get_selected_object()
        # update object property
        changed = obj.update_property(prop, value, scaling_factor)
        # update ui to display properties of selected object
        self.selection_update()

        return changed


    def get_selected_object(self):
        """always returns the whole object instance,
        even if controlpoints are selected
        """
        return self.get_full_object(self.selected_object)
    
    def get_full_object(self, obj):
        if isinstance(obj, (ControlPoints, LineControlPoints)):
            return obj.parent
        return obj

    def delete_object(self, object=None):
        # delete selected object if no other is given
        if object is None:
            object = self.selected_object
        # delete object
        if object is not None:
            self.data_handler.delete_object(object)

            if object.selected:
                self.unselect()

            self.main_ui.update_object_list()
            object.delete()

        self.main_window.right_ui.update_intensity_plot()

    def delete_object_w_undo(self, object=None):
        if object is not None or self.selected_object is not None:
            command = DeleteObjectCommand(self.data_handler, object)
            self.main_window.undo_stack.push(command)

    def set_origin(self, point:np.ndarray, move_objects=True):
        """Set the origin of the image to a given point"""
        self.data_handler.logger.debug(f"set origin to {point}")

        if move_objects:
            self.move_all_objects(-(point-self.origin))
        self.origin = point 
        self.image.transform = linear.STTransform(translate=-self.origin)
        self.center_image()

    def move_all_objects(self, vector:np.ndarray):
        """Move all objects by a relative vector"""
        self.data_handler.logger.debug("move all objects")
        for structure in self.data_handler.drawing_data.keys():
            for obj in self.data_handler.drawing_data[structure]:
                # show object
                obj.set_center(obj.center + vector)
                obj.update_from_controlpoints()
        self.scene.update()        

    def set_origin_w_undo(self, pos):
        command = SetOriginCommand(self, pos)
        self.main_window.undo_stack.push(command)

    def select(self, obj):
        self.unselect()
        self.selected_object = obj
        self.selected_object.select(True, obj=obj)
        self.selected_object.set_visibility(True)
        self.selection_update()

    def unselect(self):
        if self.selected_object is not None:
            self.selected_object.select(False)
            self.selected_object = None
            self.main_ui.clear_object_table()

    def show_all_objects(self):
        self.data_handler.logger.info("show all objects")
        for structure in self.data_handler.drawing_data.keys():
            for obj in self.data_handler.drawing_data[structure]:
                # show object
                obj.set_visibility(True)
        self.scene.update()

    def show_all_objects_w_undo(self):
        command = ShowObjectCommand(self)
        self.main_window.undo_stack.push(command)

    def hide_all_objects(self):
        self.data_handler.logger.info("hide all objects")
        for structure in self.data_handler.drawing_data.keys():
            for obj in self.data_handler.drawing_data[structure]:
                # hide object
                obj.set_visibility(False)
        self.scene.update()

    def hide_all_objects_w_undo(self):
        command = HideObjectCommand(self)
        self.main_window.undo_stack.push(command)


class DeleteObjectCommand(QUndoCommand):
    def __init__(self, data_handler, object=None):
        super().__init__()
        self.data_handler = data_handler
        self.main_window = self.data_handler.main_window
        self.vispy_canvas = self.main_window.vispy_canvas
        if object is None:
            object = self.vispy_canvas.selected_object

        object = self.vispy_canvas.get_full_object(object)

        self.object = object
        self.structure, self.index = data_handler.find_object(self.object)

    def undo(self):
        # Restore the old state
        self.data_handler.logger.info("Undoing delete object")
        if self.structure in self.data_handler.drawing_data.keys():
            self.data_handler.drawing_data[self.structure].insert(
                self.index, self.object
            )
        else:
            self.data_handler.drawing_data[self.structure] = [self.object]
        self.main_window.main_ui.update_structure_dd()
        self.object.parent = self.vispy_canvas.view.scene

        self.data_handler.logger.info("Restored object")

    def redo(self):
        # Delete object
        self.vispy_canvas.delete_object(self.object)
        self.data_handler.logger.info("Deleted object")


class MoveObjectCommand(QUndoCommand):
    def __init__(self, vispy_instance, object):
        super().__init__()
        # make a copy of the object to be able to restore the old state
        self.vispy_instance = vispy_instance

        # if object is a control point, get the whole object
        self.object = vispy_instance.get_full_object(object)

        # for redoing
        self.redoing = False
        if isinstance(self.object, (EditLineVisual, EditPolygonVisual)):
            self.coords = None
            self.old_coords = object.coords.copy()
        elif isinstance(self.object, (EditEllipseVisual, EditRectVisual)):
            
            self.center = None
            self.height = None
            self.width = None
            self.angle = None
            self.old_height = self.object.height
            self.old_width = self.object.width
            self.old_center = self.object.center
            self.old_angle = self.object.angle
        # for undoing

    def undo(self):
        self.vispy_instance.data_handler.logger.info("Undoing move object")

        # save state before undoing move
        if isinstance(self.object, (EditLineVisual, EditPolygonVisual)):
            self.object.coords = self.old_coords
        elif isinstance(self.object, (EditEllipseVisual, EditRectVisual)):

            self.object.set_center(self.old_center)
            self.object.width = self.old_width
            self.object.height = self.old_height
            self.object.angle = self.old_angle

        # restore old state

        self.object.update_from_controlpoints()
        self.vispy_instance.selection_update()

    def redo(self):
        # Move the object
        self.vispy_instance.data_handler.logger.info("Redoing move object")

        # in the first creation we don't need to set it as it is already set
        # but we need to save the state
        if not self.redoing:
            if isinstance(self.object, (EditLineVisual, EditPolygonVisual)):
                self.coords = self.object.coords.copy()
            elif isinstance(self.object, (EditEllipseVisual, EditRectVisual)):
                self.center = self.object.center.copy()
                self.height = self.object.height
                self.width = self.object.width
                self.angle = self.object.angle
            self.redoing = True
        # move object on all redo operations except the first one
        else:
            if isinstance(self.object, (EditLineVisual, EditPolygonVisual)):
                self.object.coords = self.coords
            elif isinstance(self.object, (EditEllipseVisual, EditRectVisual)):
                self.object.set_center(self.center)
                self.object.width = self.width
                self.object.height = self.height
                self.object.angle = self.angle

        self.object.update_from_controlpoints()
        self.vispy_instance.selection_update()
        
    def check_movement(self):
        if isinstance(self.object, (EditLineVisual, EditPolygonVisual)):
            if len(self.old_coords) == len(self.object.coords):
                return (self.old_coords != self.object.coords).any()
            return True
        elif isinstance(self.object, (EditEllipseVisual, EditRectVisual)):
            return (
                self.old_width != self.object.width
                or self.old_height != self.object.height
                or (self.old_center != self.object.center).any()
                or self.old_angle != self.object.angle
            )


class HideObjectCommand(QUndoCommand):
    def __init__(self, vispy_canvas):
        super().__init__()
        self.vispy_canvas = vispy_canvas

    def undo(self):
        # Show the object
        self.vispy_canvas.show_all_objects()

    def redo(self):
        # Hide the object
        self.vispy_canvas.hide_all_objects()


class ShowObjectCommand(QUndoCommand):
    def __init__(self, vispy_canvas):
        super().__init__()
        self.vispy_canvas = vispy_canvas

    def undo(self):
        # Show the object
        self.vispy_canvas.hide_all_objects()

    def redo(self):
        # Hide the object
        self.vispy_canvas.show_all_objects()


class FindScalingBarWidthCommand(QUndoCommand):
    def __init__(self, vispy_canvas, seed_points, relative, threshold, direction):
        super().__init__()
        self.vispy_canvas:VispyCanvas = vispy_canvas

        self.seed_points = seed_points
        self.relative = relative
        self.threshold = threshold
        self.direction = direction

        self.old_scale_bar_params = self.vispy_canvas.scale_bar_params

    def undo(self):
        # Restore the old state
        self.vispy_canvas.find_scale_bar_width(*self.old_scale_bar_params)

    def redo(self):
        # Find the scaling bar width
        self.vispy_canvas.find_scale_bar_width(
            self.seed_points, self.relative, self.threshold, self.direction
        )


class CreateObjectCommand(QUndoCommand):
    def __init__(self, vispy_canvas, new_object, pos, selected, structure_name):
        super().__init__()
        self.vispy_canvas = vispy_canvas
        self.new_object = self.vispy_canvas.get_full_object(new_object)
        self.pos = pos
        self.selected = selected
        self.structure_name = structure_name
        self.parent = new_object.parent

    def undo(self):
        # Delete object
        self.vispy_canvas.data_handler.logger.info("Undoing create object")
        self.vispy_canvas.delete_object(self.new_object)

    def redo(self):

        # when it is an undo->redo operation some parameters are set
        if self.new_object and self.new_object.parent is None:
            self.selected = False
            self.new_object.parent = self.parent
            self.pos = None

        # Create object
        self.structure_name = self.vispy_canvas.create_new_object(
            self.new_object, self.pos, self.selected, self.structure_name
        )


class UpdateObjectCommand(QUndoCommand):
    def __init__(self, vispy_canvas, obj, prop, value, scaling_factor, old_value):
        super().__init__()

        self.vispy_canvas = vispy_canvas
        self.obj = obj
        self.prop = prop
        self.value = value
        self.scaling_factor = scaling_factor
        self.old_value = old_value

        if scaling_factor is None:
            self.scaling_factor = 1
    
    def undo(self):
        
        # logging
        if self.scaling_factor is None:
            scaling = 1
        else:
            scaling = self.scaling_factor
        self.vispy_canvas.data_handler.logger.debug(
            f"Undoing Change {self.prop} from {self.old_value*scaling} to {self.value}"
        )

        # update object property
        self.vispy_canvas.update_object_property(
            self.obj, self.prop, self.old_value, self.scaling_factor
        )


    def redo(self):
        # logging
        if self.scaling_factor is None:
            scaling = 1
        else:
            scaling = self.scaling_factor
        self.vispy_canvas.data_handler.logger.debug(
            f"Change {self.prop} from {self.old_value*scaling} to {self.value}"
        )

        # update object property
        self.vispy_canvas.update_object_property(
            self.obj, self.prop, self.value, self.scaling_factor
        )
        
class RemovePointCommand(QUndoCommand):

    def __init__(self, vispy_canvas, object):
        super().__init__()
        self.object = object
        self.delete_index = object.control_points.index(object.selected_cp)
        self.old_point = None
        self.vispy_canvas = vispy_canvas
        self.continue_adding_points = object.continue_adding_points

    def undo(self):
        # Show the object
        self.vispy_canvas.data_handler.logger.debug(
            f"Undoing remove point at index {self.delete_index}"
        )
        self.object.continue_adding_points = True
        self.object.add_point(self.old_point, index=self.delete_index)
        self.object.continue_adding_points = self.continue_adding_points

    def redo(self):
        # Hide the object
        self.vispy_canvas.data_handler.logger.debug(
            f"Redoing remove point at index {self.delete_index}"
        )
        self.old_point = self.object.coords[self.delete_index]
        self.object.remove_point(self.delete_index)

class AddPointCommand(QUndoCommand):

    def __init__(self, vispy_canvas, object, point):
        super().__init__()
        self.object = object
        self.add_index = object.get_selected_index() + 1
        self.point = point
        self.vispy_canvas = vispy_canvas
        self.redoing = False

    def undo(self):
        # remove movable point on undo
        self.object.continue_adding_points = False
        self.vispy_canvas.data_handler.logger.debug(
            f"Undoing adding point at index {self.add_index}"
        )
        if self.add_index < len(self.object.coords):
            self.point = self.object.coords[self.add_index]
            self.object.remove_point(self.add_index)

    def redo(self):
        self.vispy_canvas.data_handler.logger.debug(
            f"Redoing adding point at index {self.add_index}"
        )
        
        show=False
        if self.object is self.vispy_canvas.selected_object:
            self.object.select(True)
            show = True

        # when redoing, we need to add the point at the index 
        # before the one we added because that one is the currently moving one 
        if self.redoing:
            self.object.add_point(self.point, index=self.add_index, select=False, show=show)
        else:
            self.object.add_point(self.point, index=self.add_index)
        self.redoing = True

class SetOriginCommand(QUndoCommand):
    def __init__(self, vispy_canvas, pos):
        super().__init__()
        self.vispy_canvas = vispy_canvas
        self.old_origin = self.vispy_canvas.origin
        self.new_origin = pos

    def undo(self):
        # Show the object
        self.vispy_canvas.set_origin(self.old_origin)

    def redo(self):
        # Hide the object
        self.vispy_canvas.set_origin(self.new_origin)